//! `NcPlane*` methods and associated functions.

use core::ptr::{null, null_mut};
use std::ffi::CStr;

use crate::{
    cstring, NcAlign, NcAlphaBits, NcBoxMask, NcCell, NcChannel, NcChannelPair, NcColor,
    NcDimension, NcEgc, NcFadeCb, NcOffset, NcPaletteIndex, NcPlane, NcPlaneOptions, NcResizeCb,
    NcResult, NcRgb, NcStyleMask, NcTime, Notcurses,
};

/// # `NcPlaneOptions` Constructors
impl NcPlaneOptions {
    /// New NcPlaneOptions using the horizontal x.
    pub fn new(y: NcOffset, x: NcOffset, rows: NcDimension, cols: NcDimension) -> Self {
        Self::with_flags(y, x, rows, cols, None, 0)
    }

    /// New NcPlaneOptions with horizontal alignment.
    pub fn new_aligned(y: NcOffset, align: NcAlign, rows: NcDimension, cols: NcDimension) -> Self {
        Self::with_flags_aligned(y, align, rows, cols, None, crate::NCPLANE_OPTION_HORALIGNED)
    }

    /// New NcPlaneOptions, with flags.
    pub fn with_flags(
        y: NcOffset,
        x: NcOffset,
        rows: NcDimension,
        cols: NcDimension,
        resizecb: Option<NcResizeCb>,
        flags: u64,
    ) -> Self {
        NcPlaneOptions {
            y: y as i32,
            x: x as i32,
            rows: rows as i32,
            cols: cols as i32,
            userptr: null_mut(),
            name: null(),
            resizecb: crate::ncresizecb_to_c(resizecb),
            flags,
        }
    }

    /// New NcPlaneOptions, with flags and horizontal alignment.
    ///
    /// Note: Already includes the
    /// [NCPLANE_OPTION_HORALIGNED][crate::NCPLANE_OPTION_HORALIGNED] flag.
    pub fn with_flags_aligned(
        y: NcOffset,
        align: NcAlign,
        rows: NcDimension,
        cols: NcDimension,
        resizecb: Option<NcResizeCb>,
        flags: u64,
    ) -> Self {
        let flags = crate::NCPLANE_OPTION_HORALIGNED | flags;
        NcPlaneOptions {
            y: y as i32,
            x: align as i32,
            rows: rows as i32,
            cols: cols as i32,
            userptr: null_mut(),
            name: null(),
            resizecb: crate::ncresizecb_to_c(resizecb),
            flags,
        }
    }
}

/// # NcPlane constructors and destructors
impl NcPlane {
    /// New NcPlane.
    ///
    /// The returned plane will be the top, bottom, and root of this new pile.
    pub fn new<'a>(
        nc: &mut Notcurses,
        y: NcOffset,
        x: NcOffset,
        rows: NcDimension,
        cols: NcDimension,
    ) -> &'a mut NcPlane {
        let options = NcPlaneOptions::new(y, x, rows, cols);
        unsafe { &mut *crate::ncpile_create(nc, &options) }
    }

    /// New NcPlane, expects an [NcPlaneOptions] struct.
    ///
    /// The returned plane will be the top, bottom, and root of this new pile.
    pub fn with_options<'a>(nc: &mut Notcurses, options: &NcPlaneOptions) -> &'a mut NcPlane {
        unsafe { &mut *crate::ncpile_create(nc, options) }
    }

    /// New NcPlane, bound to another NcPlane.
    pub fn new_bound<'a>(
        bound_to: &mut NcPlane,
        y: NcOffset,
        x: NcOffset,
        rows: NcDimension,
        cols: NcDimension,
    ) -> &'a mut NcPlane {
        let options = NcPlaneOptions::new(y, x, rows, cols);
        unsafe { &mut *crate::ncplane_create(bound_to, &options) }
    }

    /// New NcPlane, bound to another plane, expects an [NcPlaneOptions] struct.
    ///
    /// The returned plane will be the top, bottom, and root of this new pile.
    pub fn with_options_bound<'a>(
        bound_to: &mut NcPlane,
        options: &NcPlaneOptions,
    ) -> &'a mut NcPlane {
        unsafe { &mut *crate::ncplane_create(bound_to, options) }
    }

    /// New NcPlane, with the same dimensions of the terminal.
    ///
    /// The returned plane will be the top, bottom, and root of this new pile.
    pub fn new_termsize<'a>(nc: &mut Notcurses) -> &'a mut NcPlane {
        let (mut trows, mut tcols) = (0, 0);
        crate::notcurses_term_dim_yx(nc, &mut trows, &mut tcols);
        assert![(trows > 0) & (tcols > 0)];
        unsafe {
            &mut *crate::ncpile_create(
                nc,
                &NcPlaneOptions::new(0, 0, trows as NcDimension, tcols as NcDimension),
            )
        }
    }

    /// Destroys this NcPlane.
    ///
    /// None of its contents will be visible after the next render call.
    /// It is an error to attempt to destroy the standard plane.
    pub fn destroy(&mut self) -> NcResult {
        unsafe { crate::ncplane_destroy(self) }
    }
}

// -----------------------------------------------------------------------------
/// ## NcPlane methods: `NcAlphaBits`
impl NcPlane {
    /// Gets the foreground [NcAlphaBits] from this NcPlane, shifted to LSBs.
    #[inline]
    pub fn fg_alpha(&self) -> NcAlphaBits {
        crate::channels_fg_alpha(unsafe { crate::ncplane_channels(self) })
    }

    /// Gets the background [NcAlphaBits] for this NcPlane, shifted to LSBs.
    #[inline]
    pub fn bg_alpha(&self) -> NcAlphaBits {
        crate::channels_bg_alpha(unsafe { crate::ncplane_channels(self) })
    }

    /// Sets the foreground [NcAlphaBits] from this NcPlane.
    pub fn set_fg_alpha(&mut self, alpha: NcAlphaBits) -> NcResult {
        unsafe { crate::ncplane_set_fg_alpha(self, alpha as i32) }
    }

    /// Sets the background [NcAlphaBits] for this NcPlane.
    pub fn set_bg_alpha(&mut self, alpha: NcAlphaBits) -> NcResult {
        unsafe { crate::ncplane_set_bg_alpha(self, alpha as i32) }
    }
}

// -----------------------------------------------------------------------------
/// ## NcPlane methods: `NcChannel`
impl NcPlane {
    /// Gets the current [NcChannelPair] from this NcPlane.
    pub fn channels(&self) -> NcChannelPair {
        unsafe { crate::ncplane_channels(self) }
    }

    /// Gets the foreground [NcChannel] from an [NcPlane].
    #[inline]
    pub fn fchannel(&self) -> NcChannel {
        crate::channels_fchannel(unsafe { crate::ncplane_channels(self) })
    }

    /// Gets the background [NcChannel] from an [NcPlane].
    #[inline]
    pub fn bchannel(&self) -> NcChannel {
        crate::channels_bchannel(unsafe { crate::ncplane_channels(self) })
    }

    /// Sets the current [NcChannelPair] for this NcPlane.
    pub fn set_channels(&mut self, channels: NcChannelPair) {
        unsafe { crate::ncplane_set_channels(self, channels) }
    }

    /// Sets the current foreground [NcChannel] for this NcPlane.
    /// Returns the updated [NcChannelPair].
    pub fn set_fchannel(&mut self, channel: NcChannel) -> NcChannelPair {
        unsafe { crate::ncplane_set_fchannel(self, channel) }
    }

    /// Sets the current background [NcChannel] for this NcPlane.
    /// Returns the updated [NcChannelPair].
    pub fn set_bchannel(&mut self, channel: NcChannel) -> NcChannelPair {
        unsafe { crate::ncplane_set_bchannel(self, channel) }
    }

    /// Sets the given [NcChannelPair]s throughout the specified region,
    /// keeping content and attributes unchanged.
    ///
    /// Returns the number of cells set, or [NCRESULT_ER][crate::NCRESULT_ERR]
    /// on failure.
    pub fn stain(
        &mut self,
        y_stop: NcDimension,
        x_stop: NcDimension,
        ul: NcChannelPair,
        ur: NcChannelPair,
        ll: NcChannelPair,
        lr: NcChannelPair,
    ) -> NcResult {
        unsafe { crate::ncplane_stain(self, y_stop as i32, x_stop as i32, ul, ur, ll, lr) }
    }
}

// -----------------------------------------------------------------------------
/// ## NcPlane methods: `NcColor`, `NcRgb` & default color
impl NcPlane {
    /// Gets the foreground [NcColor] RGB components from this NcPlane.
    #[inline]
    pub fn fg_rgb8(&self) -> (NcColor, NcColor, NcColor) {
        let (mut r, mut g, mut b) = (0, 0, 0);
        let _ = crate::channels_fg_rgb8(
            unsafe { crate::ncplane_channels(self) },
            &mut r,
            &mut g,
            &mut b,
        );
        (r, g, b)
    }

    /// Gets the background [NcColor] RGB components from this NcPlane.
    #[inline]
    pub fn bg_rgb8(&self) -> (NcColor, NcColor, NcColor) {
        let (mut r, mut g, mut b) = (0, 0, 0);
        let _ = crate::channels_bg_rgb8(
            unsafe { crate::ncplane_channels(self) },
            &mut r,
            &mut g,
            &mut b,
        );
        (r, g, b)
    }

    /// Sets the foreground [NcColor] RGB components for this NcPlane.
    ///
    /// If the terminal does not support directly-specified 3x8b cells
    /// (24-bit "TrueColor", indicated by the "RGB" terminfo capability),
    /// the provided values will be interpreted in some lossy fashion.
    ///
    /// "HP-like" terminals require setting foreground and background at the same
    /// time using "color pairs"; Notcurses will manage color pairs transparently.
    pub fn set_fg_rgb8(&mut self, red: NcColor, green: NcColor, blue: NcColor) {
        unsafe {
            // Can't fail because of type enforcing.
            let _ = crate::ncplane_set_fg_rgb8(self, red as i32, green as i32, blue as i32);
        }
    }

    /// Sets the background [NcColor] RGB components for this NcPlane.
    ///
    /// If the terminal does not support directly-specified 3x8b cells
    /// (24-bit "TrueColor", indicated by the "RGB" terminfo capability),
    /// the provided values will be interpreted in some lossy fashion.
    ///
    /// "HP-like" terminals require setting foreground and background at the same
    /// time using "color pairs"; Notcurses will manage color pairs transparently.
    pub fn set_bg_rgb8(&mut self, red: NcColor, green: NcColor, blue: NcColor) {
        unsafe {
            // Can't fail because of type enforcing.
            let _ = crate::ncplane_set_bg_rgb8(self, red as i32, green as i32, blue as i32);
        }
    }

    /// Gets the foreground [NcRgb] from this NcPlane, shifted to LSBs.
    #[inline]
    pub fn fg_rgb(&self) -> NcRgb {
        crate::channels_fg_rgb(unsafe { crate::ncplane_channels(self) })
    }

    /// Gets the background [NcRgb] from this NcPlane, shifted to LSBs.
    #[inline]
    pub fn bg_rgb(&self) -> NcRgb {
        crate::channels_bg_rgb(unsafe { crate::ncplane_channels(self) })
    }

    /// Sets the foreground [NcRgb] for this NcPlane.
    #[inline]
    pub fn set_fg_rgb(&mut self, rgb: NcRgb) {
        unsafe {
            crate::ncplane_set_fg_rgb(self, rgb);
        }
    }

    /// Sets the backround [NcRgb] for this NcPlane.
    #[inline]
    pub fn set_bg_rgb(&mut self, rgb: NcRgb) {
        unsafe {
            crate::ncplane_set_bg_rgb(self, rgb);
        }
    }

    /// Is this NcPlane's foreground using the "default foreground color"?
    #[inline]
    pub fn fg_default_p(&self) -> bool {
        crate::channels_fg_default_p(unsafe { crate::ncplane_channels(self) })
    }

    /// Is this NcPlane's background using the "default background color"?
    #[inline]
    pub fn bg_default_p(&self) -> bool {
        crate::channels_bg_default_p(unsafe { crate::ncplane_channels(self) })
    }

    /// Uses the default color for the foreground.
    #[inline]
    pub fn set_fg_default(&mut self) {
        unsafe {
            crate::ncplane_set_fg_default(self);
        }
    }

    /// Uses the default color for the background.
    #[inline]
    pub fn set_bg_default(&mut self) {
        unsafe {
            crate::ncplane_set_bg_default(self);
        }
    }
}

// -----------------------------------------------------------------------------
/// ## NcPlane methods: `NcStyleMask` & `PaletteIndex`
impl NcPlane {
    /// Sets the given style throughout the specified region, keeping content
    /// and channels unchanged.
    ///
    /// Returns the number of cells set, or [NCRESULT_ERR][crate::NCRESULT_ERR]
    /// on failure.
    pub fn format(
        &mut self,
        y_stop: NcDimension,
        x_stop: NcDimension,
        stylemask: NcStyleMask,
    ) -> NcResult {
        unsafe { crate::ncplane_format(self, y_stop as i32, x_stop as i32, stylemask as u32) }
    }

    /// Returns the current styling for this NcPlane.
    pub fn styles(&self) -> NcStyleMask {
        unsafe { crate::ncplane_styles(self) }
    }

    /// Removes the specified styles from this NcPlane's existing spec.
    pub fn off_styles(&mut self, stylemask: NcStyleMask) {
        unsafe {
            crate::ncplane_off_styles(self, stylemask as u32);
        }
    }

    /// Adds the specified styles to this NcPlane's existing spec.
    pub fn on_styles(&mut self, stylemask: NcStyleMask) {
        unsafe {
            crate::ncplane_on_styles(self, stylemask as u32);
        }
    }

    /// Sets just the specified styles for this NcPlane.
    pub fn set_styles(&mut self, stylemask: NcStyleMask) {
        unsafe {
            crate::ncplane_set_styles(self, stylemask as u32);
        }
    }

    /// Sets this NcPlane's foreground [NcPaletteIndex].
    ///
    /// Also sets the foreground palette index bit, sets it foreground-opaque,
    /// and clears the foreground default color bit.
    pub fn set_fg_palindex(&mut self, palindex: NcPaletteIndex) {
        unsafe {
            crate::ncplane_set_fg_palindex(self, palindex as i32);
        }
    }

    /// Sets this NcPlane's background [NcPaletteIndex].
    ///
    /// Also sets the background palette index bit, sets it background-opaque,
    /// and clears the background default color bit.
    pub fn set_bg_palindex(&mut self, palindex: NcPaletteIndex) {
        unsafe {
            crate::ncplane_set_bg_palindex(self, palindex as i32);
        }
    }
}

// -----------------------------------------------------------------------------
/// ## NcPlane methods: `NcCell` & `NcEgc`
impl NcPlane {
    /// Retrieves the current contents of the [NcCell] under the cursor,
    /// returning the [NcEgc] and writing out the [NcStyleMask] and the [NcChannelPair].
    ///
    /// This NcEgc must be freed by the caller.
    pub fn at_cursor(
        &mut self,
        stylemask: &mut NcStyleMask,
        channels: &mut NcChannelPair,
    ) -> Option<NcEgc> {
        let egc = unsafe { crate::ncplane_at_cursor(self, stylemask, channels) };
        if egc.is_null() {
            return None;
        }
        let egc = core::char::from_u32(unsafe { *egc } as u32).expect("wrong char");
        Some(egc)
    }

    /// Retrieves the current contents of the [NcCell] under the cursor.
    ///
    /// This NcCell is invalidated if the associated NcPlane is destroyed.
    #[inline]
    pub fn at_cursor_cell(&mut self, cell: &mut NcCell) -> NcResult {
        crate::ncplane_at_cursor_cell(self, cell)
    }

    /// Retrieves the current contents of the specified [NcCell], returning the
    /// [NcEgc] and writing out the [NcStyleMask] and the [NcChannelPair].
    ///
    /// This NcEgc must be freed by the caller.
    pub fn at_yx(
        &mut self,
        y: NcDimension,
        x: NcDimension,
        stylemask: &mut NcStyleMask,
        channels: &mut NcChannelPair,
    ) -> Option<NcEgc> {
        let egc = unsafe { crate::ncplane_at_yx(self, y as i32, x as i32, stylemask, channels) };
        if egc.is_null() {
            return None;
        }
        let egc = core::char::from_u32(unsafe { *egc } as u32).expect("wrong char");
        Some(egc)
    }

    /// Retrieves the current contents of the specified [NcCell] into `cell`.
    /// This cell is invalidated if the associated plane is destroyed.
    #[inline]
    pub fn ncplane_at_yx_cell(
        &mut self,
        y: NcDimension,
        x: NcDimension,
        cell: &mut NcCell,
    ) -> NcResult {
        crate::ncplane_at_yx_cell(self, y, x, cell)
    }

    /// Extracts this NcPlane's base [NcCell] into `cell`.
    ///
    /// The reference is invalidated if this NcPlane is destroyed.
    pub fn base(&mut self, cell: &mut NcCell) -> NcResult {
        unsafe { crate::ncplane_base(self, cell) }
    }

    /// Sets this NcPlane's base NcCell from its components.
    ///
    /// This function must be called with an empty `egc`.
    /// `egc` must be a single extended grapheme cluster.
    ///
    /// It will be used for purposes of rendering anywhere that the NcPlane's
    /// gcluster is 0.
    ///
    /// Erasing the NcPlane does not reset the base cell.
    pub fn set_base(
        &mut self,
        egc: &NcEgc,
        stylemask: NcStyleMask,
        channels: NcChannelPair,
    ) -> NcResult {
        unsafe { crate::ncplane_set_base(self, &(*egc as i8), stylemask as u32, channels) }
    }

    /// Sets this NcPlane's base NcCell.
    ///
    /// This function must be called with a zero `cell`.
    ///
    /// It will be used for purposes of rendering anywhere that the NcPlane's
    /// gcluster is 0.
    ///
    /// Erasing the NcPlane does not reset the base cell.
    // FIXME: documentation: https://github.com/dankamongmen/notcurses/issues/1238
    pub fn set_base_cell(&mut self, cell: &NcCell) -> NcResult {
        unsafe { crate::ncplane_set_base_cell(self, cell) }
    }

    /// Creates a flat string from the NcEgc's of the selected region of the
    /// NcPlane.
    ///
    /// Starts at the plane's `beg_y` * `beg_x` coordinates (which must lie on
    /// the plane), continuing for `len_y` x `len_x` cells.
    ///
    /// If either `through_y` or `through_x` are true, then `len_y` or `len_x`,
    /// will be respectively ignored, and will go through the boundary of the plane.
    pub fn contents(
        &self,
        beg_y: NcDimension,
        beg_x: NcDimension,
        len_y: NcDimension,
        len_x: NcDimension,
        through_y: bool,
        through_x: bool,
    ) -> String {
        let (mut len_y, mut len_x) = (len_y as i32, len_x as i32);
        if through_y {
            len_y = -1;
        }
        if through_x {
            len_x = -1;
        }
        unsafe {
            CStr::from_ptr(crate::ncplane_contents(
                self,
                beg_y as i32,
                beg_x as i32,
                len_y,
                len_x,
            ))
            .to_string_lossy()
            .into_owned()
        }
    }

    /// Erases every NcCell in this NcPlane, resetting all attributes to normal,
    /// all colors to the default color, and all cells to undrawn.
    ///
    /// All cells associated with this NcPlane are invalidated, and must not be
    /// used after the call, excluding the base cell. The cursor is homed.
    pub fn erase(&mut self) {
        unsafe { crate::ncplane_erase(self) }
    }

    /// Replaces the NcCell at the specified coordinates with the provided NcCell,
    /// advancing the cursor by its width (but not past the end of the plane).
    ///
    /// The new NcCell must already be associated with the Plane.
    /// On success, returns the number of columns the cursor was advanced.
    /// On failure, -1 is returned.
    pub fn putc_yx(&mut self, y: NcDimension, x: NcDimension, cell: &NcCell) -> NcResult {
        unsafe { crate::ncplane_putc_yx(self, y as i32, x as i32, cell) }
    }

    /// Replaces the NcCell at the current coordinates with the provided NcCell,
    /// advancing the cursor by its width (but not past the end of the plane).
    ///
    /// The new NcCell must already be associated with the Plane.
    /// On success, returns the number of columns the cursor was advanced.
    pub fn putc(&mut self, cell: &NcCell) -> NcResult {
        crate::ncplane_putc(self, cell)
    }

    /// Calls [putchar_yx](type.NcPlane.html#method.putchar_yx) at the current cursor location.
    pub fn putchar(&mut self, ch: char) -> NcResult {
        crate::ncplane_putchar(self, ch)
    }

    // TODO: call put_egc
    // /// Replaces the [NcEgc][crate::NcEgc] to the current location, but retain
    // /// the styling. The current styling of the plane will not be changed.
    // pub fn putchar_stained(&mut self, y: NcDimension, x: NcDimension, ch: char) -> NcResult {
    //     crate::ncplane_putchar_stained(self, ch)
    // }

    /// Replaces the [NcEgc][crate::NcEgc], but retain the styling.
    /// The current styling of the plane will not be changed.
    pub fn putchar_yx(&mut self, y: NcDimension, x: NcDimension, ch: char) -> NcResult {
        crate::ncplane_putchar_yx(self, y, x, ch)
    }

    /// Writes a series of [NcEgc][crate::NcEgc]s to the current location,
    /// using the current style.
    ///
    /// Advances the cursor by some positive number of columns (though not beyond
    /// the end of the plane); this number is returned on success.
    ///
    /// On error, a non-positive number is returned, indicating the number of
    /// columns which were written before the error.
    #[inline]
    pub fn putstr(&mut self, string: &str) -> NcResult {
        crate::ncplane_putstr(self, string)
    }

    /// Writes a series of [NcEgc][crate::NcEgc]s to the current location, but
    /// retain the styling.
    /// The current styling of the plane will not be changed.
    pub fn putstr_stained(&mut self, string: &str) -> NcResult {
        unsafe { crate::ncplane_putstr_stained(self, cstring![string]) }
    }

    /// Write a string, which is a series of [NcEgc][crate::NcEgc]s, to the
    /// current location, using the current style.
    ///
    /// They will be interpreted as a series of columns.
    ///
    /// Advances the cursor by some positive number of columns (though not
    /// beyond the end of the plane); this number is returned on success.
    ///
    /// On error, a non-positive number is returned, indicating the number of
    /// columns which were written before the error.
    pub fn putstr_yx(&mut self, y: NcDimension, x: NcDimension, string: &str) -> NcResult {
        unsafe { crate::ncplane_putstr_yx(self, y as i32, x as i32, cstring![string]) }
    }
}

// -----------------------------------------------------------------------------
/// ## NcPlane methods: `NcPlane` & `Notcurses`
impl NcPlane {
    /// Duplicates this NcPlane.
    ///
    /// The new NcPlane will have the same geometry, the same rendering state,
    /// and all the same duplicated content.
    ///
    /// The new plane will be immediately above the old one on the z axis,
    /// and will be bound to the same parent. Bound planes are not duplicated;
    /// the new plane is bound to the current parent, but has no bound planes.
    // TODO: deal with the opaque field.
    pub fn dup<'a>(&'a mut self) -> &'a mut NcPlane {
        unsafe { &mut *crate::ncplane_dup(self, null_mut()) }
    }

    /// Returns the topmost NcPlane of the current pile.
    pub fn top<'a>(&mut self) -> &'a mut NcPlane {
        unsafe { &mut *crate::ncpile_top(self) }
    }

    /// Returns the bottommost NcPlane of the current pile.
    pub fn bottom<'a>(&mut self) -> &'a mut NcPlane {
        unsafe { &mut *crate::ncpile_bottom(self) }
    }

    /// Relocates this NcPlane at the top of the z-buffer.
    pub fn move_top(&mut self) {
        unsafe {
            crate::ncplane_move_top(self);
        }
    }

    /// Relocates this NcPlane at the bottom of the z-buffer.
    pub fn move_bottom(&mut self) {
        unsafe {
            crate::ncplane_move_bottom(self);
        }
    }

    /// Moves this NcPlane relative to the standard plane, or the plane to
    /// which it is bound.
    ///
    /// It is an error to attempt to move the standard plane.
    // CHECK: whether a negative offset is valid
    pub fn move_yx(&mut self, y: NcOffset, x: NcOffset) -> NcResult {
        unsafe { crate::ncplane_move_yx(self, y, x) }
    }

    /// Returns the NcPlane above this one, or None if already at the top.
    pub fn above<'a>(&'a mut self) -> Option<&'a mut NcPlane> {
        let plane = unsafe { crate::ncplane_above(self) };
        if plane.is_null() {
            return None;
        }
        Some(unsafe { &mut *plane })
    }

    /// Returns the NcPlane below this one, or None if already at the bottom.
    pub fn below<'a>(&'a mut self) -> Option<&'a mut NcPlane> {
        let plane = unsafe { crate::ncplane_below(self) };
        if plane.is_null() {
            return None;
        }
        Some(unsafe { &mut *plane })
    }

    /// Relocates this NcPlane above the `above` NcPlane, in the z-buffer.
    ///
    /// Returns [NCRESULT_ERR][crate::NCRESULT_ERR] if the current plane is
    /// already in the desired location. Both planes must not be the same.
    pub fn move_above(&mut self, above: &mut NcPlane) -> NcResult {
        unsafe { crate::ncplane_move_above(self, above) }
    }

    /// Relocates this NcPlane below the `below` NcPlane, in the z-buffer.
    ///
    /// Returns [NCRESULT_ERR][crate::NCRESULT_ERR] if the current plane is
    /// already in the desired location. Both planes must not be the same.
    pub fn move_below(&mut self, below: &mut NcPlane) -> NcResult {
        unsafe { crate::ncplane_move_below(self, below) }
    }

    /// Merges `source` down onto this NcPlane.
    ///
    /// Merging is independent of the position of both NcPlanes on the z-axis.
    ///
    /// It is an error to define a subregion of zero area, or that is not
    /// entirely contained within `source`.
    ///
    /// It is an error to define a target origin such that the projected
    /// subregion is not entirely contained within 'dst'.
    ///
    /// Behavior is undefined if both NcPlanes are equivalent.
    pub fn mergedown(
        &mut self,
        source: &NcPlane,
        source_y: NcDimension,
        source_x: NcDimension,
        len_y: NcDimension,
        len_x: NcDimension,
        target_y: NcDimension,
        target_x: NcDimension,
    ) -> NcResult {
        unsafe {
            crate::ncplane_mergedown(
                source,
                self,
                source_y as i32,
                source_x as i32,
                len_y as i32,
                len_x as i32,
                target_y as i32,
                target_x as i32,
            )
        }
    }

    /// Merges `source` down onto this NcPlane.
    ///
    /// If `source` does not intersect, this plane will not be changed,
    /// but it is not an error.
    ///
    /// See [`mergedown`](type.NcPlane.html#method.mergedown)
    /// for more information.
    //
    // TODO: maybe create a reversed method, and/or an associated function,
    // for `mergedown` too.
    pub fn mergedown_simple(&mut self, source: &NcPlane) -> NcResult {
        unsafe { crate::ncplane_mergedown_simple(source, self) }
    }

    /// Rotate the plane π/2 radians clockwise.
    ///
    /// This cannot be performed on arbitrary planes, because glyphs cannot be
    /// arbitrarily rotated.
    ///
    /// The glyphs which can be rotated are limited: line-drawing characters,
    /// spaces, half blocks, and full blocks.
    ///
    /// The plane must have an even number of columns.
    ///
    /// Use the ncvisual rotation for a more flexible approach.
    pub fn rotate_cw(&mut self) -> NcResult {
        unsafe { crate::ncplane_rotate_cw(self) }
    }

    /// Rotate the plane π/2 radians counter-clockwise.
    ///
    /// See [`routate_cw`](type.NcPlane.html#method.rotate_cw)
    /// for more information.
    pub fn rotate_ccw(&mut self) -> NcResult {
        unsafe { crate::ncplane_rotate_ccw(self) }
    }

    /// Gets the parent to which this NcPlane is bound, if any.
    //
    // TODO: CHECK: what happens when it's bound to itself.
    // pub fn parent<'a>(&'a mut self) -> Option<&'a mut NcPlane> {
    pub fn parent<'a>(&'a mut self) -> &'a mut NcPlane {
        unsafe { &mut *crate::ncplane_parent(self) }
    }

    /// Gets the parent to which this NcPlane is bound, if any.
    //
    // TODO: CHECK: what happens when it's bound to itself.
    // pub fn parent<'a>(&'a mut self) -> Option<&'a mut NcPlane> {
    pub fn parent_const<'a>(&'a self) -> &'a NcPlane {
        unsafe { &*crate::ncplane_parent_const(self) }
    }

    /// Unbounds this NcPlane from its parent, makes it a bound child of
    /// 'newparent', and returns itself.
    ///
    /// Any planes bound to this NcPlane are reparented to the previous parent.
    ///
    /// If this NcPlane is equal to `newparent`, then becomes the root of a new
    /// pile, unless it is already the root of a pile, in which case this is a
    /// no-op.
    ///
    /// The standard plane cannot be reparented.
    pub fn reparent<'a>(&'a mut self, newparent: &'a mut NcPlane) -> &'a mut NcPlane {
        unsafe { &mut *crate::ncplane_reparent(self, newparent) }
    }

    /// Like [`reparent`](type.NcPlane.html#method.reparent), except any bound
    /// planes comes along with this NcPlane to its new destination.
    ///
    /// Their z-order is maintained.
    ///
    // TODO:CHECK: If 'newparent' is an ancestor, NULL is returned & no changes're made.
    pub fn reparent_family<'a>(&'a mut self, newparent: &'a mut NcPlane) -> &'a mut NcPlane {
        unsafe { &mut *crate::ncplane_reparent_family(self, newparent) }
    }

    /// Makes the physical screen match the last rendered frame from the pile of
    /// which this NcPlane is a part.
    ///
    /// This is a blocking call. Don't call this before the pile has been
    /// rendered (doing so will likely result in a blank screen).
    pub fn rasterize<'a>(&mut self) -> NcResult {
        unsafe { crate::ncpile_rasterize(self) }
    }

    /// Renders the pile of which this NcPlane is a part.
    /// Rendering this pile again will blow away the render.
    /// To actually write out the render, call ncpile_rasterize().
    pub fn render<'a>(&mut self) -> NcResult {
        unsafe { crate::ncpile_render(self) }
    }

    /// Gets a mutable reference to the [Notcurses] context of this NcPlane.
    pub fn notcurses<'a>(&mut self) -> &'a mut Notcurses {
        unsafe { &mut *crate::ncplane_notcurses(self) }
    }

    /// Gets an immutable reference to the [Notcurses] context of this NcPlane.
    pub fn notcurses_const<'a>(&self) -> &'a Notcurses {
        unsafe { &*crate::ncplane_notcurses_const(self) }
    }
}

// -----------------------------------------------------------------------------
/// ## NcPlane methods: cursor
impl NcPlane {
    /// Moves the cursor to 0, 0.
    pub fn home(&mut self) {
        unsafe {
            crate::ncplane_home(self);
        }
    }

    /// Returns the current position of the cursor within this NcPlane.
    ///
    // NOTE: y and/or x may be NULL.
    // maybe check for null and return Some() or None?
    pub fn cursor_yx(&self) -> (NcDimension, NcDimension) {
        let (mut y, mut x) = (0, 0);
        unsafe { crate::ncplane_cursor_yx(self, &mut y, &mut x) };
        (y as NcDimension, x as NcDimension)
    }

    /// Returns the current row of the cursor within this NcPlane.
    pub fn cursor_y(&self) -> NcDimension {
        self.cursor_yx().0
    }

    /// Returns the current column of the cursor within this NcPlane.
    pub fn cursor_x(&self) -> NcDimension {
        self.cursor_yx().1
    }

    /// Moves the cursor to the specified position within this NcPlane.
    ///
    /// The cursor doesn't need to be visible.
    ///
    /// Parameters exceeding the plane's dimensions will result in an error,
    /// and the cursor position will remain unchanged.
    pub fn cursor_move_yx(&mut self, y: NcDimension, x: NcDimension) -> NcResult {
        unsafe { crate::ncplane_cursor_move_yx(self, y as i32, x as i32) }
    }

    /// Moves the cursor the number of rows specified (forward or backwards).
    ///
    /// It will error if the target row exceeds the plane dimensions.
    pub fn cursor_move_rows(&mut self, rows: NcOffset) -> NcResult {
        let (y, x) = self.cursor_yx();
        self.cursor_move_yx((y as NcOffset + rows) as NcDimension, x)
    }

    /// Moves the cursor the number of columns specified (forward or backwards).
    ///
    /// It will error if the target column exceeds the plane dimensions.
    // TODO: maybe in this case it can improve
    pub fn cursor_move_cols(&mut self, cols: NcOffset) -> NcResult {
        let (y, x) = self.cursor_yx();
        self.cursor_move_yx(y, (x as NcOffset + cols) as NcDimension)
    }
}

// -----------------------------------------------------------------------------
/// ## NcPlane methods: size, position & alignment
impl NcPlane {
    /// Returns the column at which `cols` columns ought start in order to be
    /// aligned according to `align` within this NcPlane.
    ///
    /// Returns `-`[NCRESULT_MAX][crate::NCRESULT_MAX] if
    /// [NCALIGN_UNALIGNED][crate::NCALIGN_UNALIGNED] or invalid [NcAlign].
    #[inline]
    pub fn align(&mut self, align: NcAlign, cols: NcDimension) -> NcResult {
        crate::ncplane_align(self, align, cols)
    }

    ///
    // TODO: doc.
    pub fn center_abs(&self, y: &mut NcDimension, x: &mut NcDimension) {
        unsafe {
            crate::ncplane_center_abs(self, &mut (*y as i32), &mut (*x as i32));
        }
    }

    /// Returns the dimensions of this NcPlane.
    pub fn dim_yx(&self) -> (NcDimension, NcDimension) {
        let (mut y, mut x) = (0, 0);
        unsafe { crate::ncplane_dim_yx(self, &mut y, &mut x) };
        (y as NcDimension, x as NcDimension)
    }

    /// Return the rows of this NcPlane.
    #[inline]
    pub fn dim_y(&self) -> NcDimension {
        self.dim_yx().0
    }

    /// Return the columns of this NcPlane.
    #[inline]
    pub fn dim_x(&self) -> NcDimension {
        self.dim_yx().1
    }

    /// Return the rows of this NcPlane.
    ///
    /// Alias of [dim_y](type.NcPlane.html#method.dim_y)
    #[inline]
    pub fn rows(&self) -> NcDimension {
        self.dim_yx().0
    }

    /// Return the cols of this NcPlane.
    ///
    /// Alias of [dim_x](type.NcPlane.html#method.dim_x)
    #[inline]
    pub fn cols(&self) -> NcDimension {
        self.dim_yx().1
    }

    /// Resizes this NcPlane.
    ///
    /// The four parameters `keep_y`, `keep_x`, `keep_len_y`, and `keep_len_x`
    /// defines a subset of this NcPlane to keep unchanged. This may be a section
    /// of size 0.
    ///
    /// `keep_x` and `keep_y` are relative to this NcPlane. They must specify a
    /// coordinate within the ncplane's totality. If either of `keep_len_y` or
    /// `keep_len_x` is non-zero, both must be non-zero.
    ///
    /// `y_off` and `x_off` are relative to `keep_y` and `keep_x`, and place the
    /// upper-left corner of the resized NcPlane.
    ///
    /// `y_len` and `x_len` are the dimensions of this NcPlane after resizing.
    /// `y_len` must be greater than or equal to `keep_len_y`,
    /// and `x_len` must be greater than or equal to `keeplenx`.
    ///
    /// It is an error to attempt to resize the standard plane.
    pub fn resize(
        &mut self,
        keep_y: NcDimension,
        keep_x: NcDimension,
        keep_len_y: NcDimension,
        keep_len_x: NcDimension,
        y_off: NcOffset,
        x_off: NcOffset,
        y_len: NcDimension,
        x_len: NcDimension,
    ) -> NcResult {
        unsafe {
            crate::ncplane_resize(
                self,
                keep_y as i32,
                keep_x as i32,
                keep_len_y as i32,
                keep_len_x as i32,
                y_off as i32,
                x_off as i32,
                y_len as i32,
                x_len as i32,
            )
        }
    }

    /// Realigns this NcPlane against its parent, using the alignment specified
    /// at creation time.
    ///
    /// Suitable for use as an [NcResizeCb].
    pub fn resize_realign(&mut self) -> NcResult {
        unsafe { crate::ncplane_resize_realign(self) }
    }

    /// Resizes this NcPlane, retaining what data we can (everything, unless we're
    /// shrinking in some dimension). Keeps the origin where it is.
    #[inline]
    pub fn resize_simple(&mut self, y_len: NcDimension, x_len: NcDimension) -> NcResult {
        crate::ncplane_resize_simple(self, y_len as u32, x_len as u32)
    }

    /// Returns this NcPlane's current resize callback.
    pub fn resizecb(&self) -> Option<NcResizeCb> {
        unsafe { crate::ncresizecb_to_rust(crate::ncplane_resizecb(self)) }
    }

    /// Replaces this NcPlane's existing resize callback (which may be [None]).
    ///
    /// The standard plane's resizecb may not be changed.
    pub fn set_resizecb(&mut self, resizecb: Option<NcResizeCb>) {
        unsafe { crate::ncplane_set_resizecb(self, crate::ncresizecb_to_c(resizecb)) }
    }

    /// Maps the provided coordinates relative to the origin of this NcPlane,
    /// to the same absolute coordinates relative to the origin of `target`.
    // TODO: API change, return the coordinates instead of using mutable references
    pub fn translate(&self, target: &NcPlane, y: &mut NcDimension, x: &mut NcDimension) {
        unsafe { crate::ncplane_translate(self, target, &mut (*y as i32), &mut (*x as i32)) }
    }

    /// If the provided absolute `y`/`x` coordinates are within this NcPlane
    /// returns true, if not returns false.
    ///
    /// Either way, translates the absolute coordinates relative to this NcPlane.
    pub fn translate_abs(&self, y: &mut NcDimension, x: &mut NcDimension) -> bool {
        unsafe { crate::ncplane_translate_abs(self, &mut (*y as i32), &mut (*x as i32)) }
    }

    /// Gets the `y`, `x` origin of this NcPlane relative to the standard plane,
    /// or the NcPlane to which it is bound.
    // CHECK: negative offsets
    pub fn yx(&self) -> (NcOffset, NcOffset) {
        let (mut y, mut x) = (0, 0);
        unsafe { crate::ncplane_yx(self, &mut y, &mut x) };
        (y as NcOffset, x as NcOffset)
    }

    /// Gets the `x` origin of this NcPlane relative to the standard plane,
    /// or the NcPlane to which it is bound.
    pub fn x(&self) -> NcOffset {
        unsafe { crate::ncplane_x(self) as NcOffset }
    }

    /// Gets the `y` origin of this NcPlane relative to the standard plane,
    /// or the NcPlane to which it is bound.
    pub fn y(&self) -> NcOffset {
        unsafe { crate::ncplane_y(self) as NcOffset }
    }

    /// Sets the scrolling behaviour of the plane, and
    /// returns true if scrolling was previously enabled, of false, if disabled.
    ///
    /// All planes are created with scrolling disabled. Attempting to print past
    /// the end of a line will stop at the plane boundary, and indicate an error.
    ///
    /// On a plane 10 columns wide and two rows high, printing "0123456789"
    /// at the origin should succeed, but printing "01234567890" will by default
    /// fail at the eleventh character. In either case, the cursor will be left
    /// at location 0x10; it must be moved before further printing can take place. I
    pub fn set_scrolling(&mut self, scroll: bool) -> bool {
        unsafe { crate::ncplane_set_scrolling(self, scroll) }
    }
}

// -----------------------------------------------------------------------------
/// ## NcPlane methods: boxes & perimeters
impl NcPlane {
    /// Draws a box with its upper-left corner at the current cursor position,
    /// and its lower-right corner at `y_stop` * `x_stop`.
    ///
    /// The 6 cells provided are used to draw the upper-left, ur, ll, and lr corners,
    /// then the horizontal and vertical lines.
    ///
    /// See [NcBoxMask] for information about the border and gradient masks,
    /// and the drawing of corners.
    ///
    /// If the gradient bit is not set, the styling from the hline/vlline cells
    /// is used for the horizontal and vertical lines, respectively.
    ///
    /// If the gradient bit is set, the color is linearly interpolated between
    /// the two relevant corner cells.
    pub fn r#box(
        &mut self,
        ul: &NcCell,
        ur: &NcCell,
        ll: &NcCell,
        lr: &NcCell,
        hline: &NcCell,
        vline: &NcCell,
        y_stop: NcDimension,
        x_stop: NcDimension,
        boxmask: NcBoxMask,
    ) -> NcResult {
        unsafe {
            crate::ncplane_box(
                self,
                ul,
                ur,
                ll,
                lr,
                hline,
                vline,
                y_stop as i32,
                x_stop as i32,
                boxmask,
            )
        }
    }

    /// Draws a box with its upper-left corner at the current cursor position,
    /// having dimensions `y_len` * `x_len`.
    /// The minimum box size is 2x2, and it cannot be drawn off-screen.
    ///
    /// See the [`box`](type.NcPlane.html#method.box) method for more information.
    #[inline]
    pub fn box_sized(
        &mut self,
        ul: &NcCell,
        ur: &NcCell,
        ll: &NcCell,
        lr: &NcCell,
        hline: &NcCell,
        vline: &NcCell,
        y_len: NcDimension,
        x_len: NcDimension,
        boxmask: NcBoxMask,
    ) -> NcResult {
        crate::ncplane_box_sized(self, ul, ur, ll, lr, hline, vline, y_len, x_len, boxmask)
    }

    ///
    #[inline]
    pub fn double_box(
        &mut self,
        stylemask: NcStyleMask,
        channels: NcChannelPair,
        y_stop: NcDimension,
        x_stop: NcDimension,
        boxmask: NcBoxMask,
    ) -> NcResult {
        crate::ncplane_double_box(self, stylemask, channels, y_stop, x_stop, boxmask)
    }

    ///
    #[inline]
    pub fn double_box_sized(
        &mut self,
        stylemask: NcStyleMask,
        channels: NcChannelPair,
        y_len: NcDimension,
        x_len: NcDimension,
        boxmask: NcBoxMask,
    ) -> NcResult {
        crate::ncplane_double_box(self, stylemask, channels, y_len, x_len, boxmask)
    }

    /// Draws the perimeter around this NcPlane.
    #[inline]
    pub fn perimeter(
        &mut self,
        ul: &NcCell,
        ur: &NcCell,
        ll: &NcCell,
        lr: &NcCell,
        hline: &NcCell,
        vline: &NcCell,
        boxmask: NcBoxMask,
    ) -> NcResult {
        crate::ncplane_perimeter(self, ul, ur, ll, lr, hline, vline, boxmask)
    }

    ///
    #[inline]
    pub fn perimeter_double(
        &mut self,
        stylemask: NcStyleMask,
        channels: NcChannelPair,
        boxmask: NcBoxMask,
    ) -> NcResult {
        crate::ncplane_perimeter_double(self, stylemask, channels, boxmask)
    }

    ///
    #[inline]
    pub fn perimeter_rounded(
        &mut self,
        stylemask: NcStyleMask,
        channels: NcChannelPair,
        boxmask: NcBoxMask,
    ) -> NcResult {
        crate::ncplane_perimeter_rounded(self, stylemask, channels, boxmask)
    }
}

// -----------------------------------------------------------------------------
/// ## NcPlane methods: fading, gradients & greyscale
impl NcPlane {
    /// Fades this NcPlane in, over the specified time, calling 'fader' at
    /// each iteration.
    ///
    /// Usage:
    /// 1. Load this NcPlane with the target cells without rendering.
    /// 2. call this function.
    ///
    /// When it's done, the NcPlane will have reached the target levels,
    /// starting from zeroes.
    pub fn fadein(&mut self, time: &NcTime, fader: NcFadeCb) -> NcResult {
        unsafe { crate::ncplane_fadein(self, time, fader, null_mut()) }
    }

    /// Fades in through 'iter' iterations,
    /// where 'iter' < 'ncfadectx_iterations(nctx)'.
    pub fn fadein_iteration(&mut self, time: &NcTime, fader: NcFadeCb) -> NcResult {
        unsafe { crate::ncplane_fadein(self, time, fader, null_mut()) }
    }

    /// Fades this NcPlane out, over the specified time, calling 'fader' at
    /// each iteration.
    ///
    /// Requires a terminal which supports truecolor, or at least palette
    /// modification (if the terminal uses a palette, our ability to fade planes
    /// is limited, and affected by the complexity of the rest of the screen).
    pub fn fadeout(&mut self, time: &NcTime, fader: NcFadeCb) -> NcResult {
        unsafe { crate::ncplane_fadeout(self, time, fader, null_mut()) }
    }

    /// Fades out through 'iter' iterations,
    /// where 'iter' < 'ncfadectx_iterations(nctx)'.
    pub fn fadeout_iteration(&mut self, time: &NcTime, fader: NcFadeCb) -> NcResult {
        unsafe { crate::ncplane_fadeout(self, time, fader, null_mut()) }
    }

    /// Pulses this NcPlane in and out until the callback returns non-zero,
    /// relying on the callback 'fader' to initiate rendering.
    ///
    /// `time` defines the half-period (i.e. the transition from black to full
    /// brightness, or back again).
    ///
    /// Proper use involves preparing (but not rendering) the NcPlane,
    /// then calling this method, which will fade in from black to the
    /// specified colors.
    pub fn pulse(&mut self, time: &NcTime, fader: NcFadeCb) -> NcResult {
        unsafe { crate::ncplane_pulse(self, time, fader, null_mut()) }
    }

    /// Draws a gradient with its upper-left corner at the current cursor
    /// position, stopping at `y_stop` * `xstop`.
    ///
    /// Returns the number of cells filled on success,
    /// or [NCRESULT_ERR][crate::NCRESULT_ERR] on error.
    ///
    /// The glyph composed of `egc` and `stylemask` is used for all cells.
    /// The channels specified by `ul`, `ur`, `ll`, and `lr` are composed into
    /// foreground and background gradients.
    ///
    /// To do a vertical gradient, `ul` ought equal `ur` and `ll` ought equal
    /// `lr`. To do a horizontal gradient, `ul` ought equal `ll` and `ur` ought
    /// equal `ul`.
    ///
    /// To color everything the same, all four channels should be equivalent.
    /// The resulting alpha values are equal to incoming alpha values.
    ///
    /// Palette-indexed color is not supported.
    ///
    /// Preconditions for gradient operations (error otherwise):
    ///
    /// all: only RGB colors, unless all four channels match as default
    /// all: all alpha values must be the same
    /// 1x1: all four colors must be the same
    /// 1xN: both top and both bottom colors must be the same (vertical gradient)
    /// Nx1: both left and both right colors must be the same (horizontal gradient)
    // TODO: https://github.com/dankamongmen/notcurses/issues/1235
    pub fn gradient(
        &mut self,
        egc: &NcEgc,
        stylemask: NcStyleMask,
        ul: NcChannel,
        ur: NcChannel,
        ll: NcChannel,
        lr: NcChannel,
        y_stop: NcDimension,
        x_stop: NcDimension,
    ) -> NcResult {
        unsafe {
            crate::ncplane_gradient(
                self,
                &(*egc as i8),
                stylemask as u32,
                ul as u64,
                ur as u64,
                ll as u64,
                lr as u64,
                y_stop as i32,
                x_stop as i32,
            )
        }
    }

    /// Draw a gradient with its upper-left corner at the current cursor position,
    /// having dimensions `y_len` * `x_len`. See ncplane_gradient for more information.
    /// static inline int
    #[inline]
    pub fn ncplane_gradient_sized(
        &mut self,
        egc: &NcEgc,
        stylemask: NcStyleMask,
        ul: NcChannel,
        ur: NcChannel,
        ll: NcChannel,
        lr: NcChannel,
        y_len: NcDimension,
        x_len: NcDimension,
    ) -> NcResult {
        crate::ncplane_gradient_sized(self, egc, stylemask, ul, ur, ll, lr, y_len, x_len)
    }

    /// Draws a high-resolution gradient using upper blocks and synced backgrounds.
    ///
    /// Returns the number of cells filled on success,
    /// or [NCRESULT_ERR][crate::NCRESULT_ERR] on error.
    ///
    /// This doubles the number of vertical gradations, but restricts you to
    /// half blocks (appearing to be full blocks).
    pub fn highgradient(
        &mut self,
        ul: NcChannel,
        ur: NcChannel,
        ll: NcChannel,
        lr: NcChannel,
        y_stop: NcDimension,
        x_stop: NcDimension,
    ) -> NcResult {
        unsafe { crate::ncplane_highgradient(self, ul, ur, ll, lr, y_stop as i32, x_stop as i32) }
    }

    /// [`gradent_sized`](type.NcPlane.html#method.gradient_sized)`()`
    /// meets [`highgradient`](type.NcPlane.html#method.highgradient)`()`.
    pub fn highgradient_sized(
        &mut self,
        ul: NcChannel,
        ur: NcChannel,
        ll: NcChannel,
        lr: NcChannel,
        y_stop: NcDimension,
        x_stop: NcDimension,
    ) -> NcResult {
        unsafe {
            crate::ncplane_highgradient_sized(self, ul, ur, ll, lr, y_stop as i32, x_stop as i32)
        }
    }

    /// Converts this NcPlane's content to greyscale.
    pub fn greyscale(&mut self) {
        unsafe {
            crate::ncplane_greyscale(self);
        }
    }
}
